/*
 **** Hash Table ****
 *
 * This file contains routines dealing with hash tables
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "datademo.h"

struct HashTable HashTable[10];	/* hash table */

/* function prototypes */
void add_hash_table(void);
void delete_hash_table(void);
void list_hash_table(void);
void search_hash_table(void);
void display_hash_table(int hashval, int ent);
int get_hashval(int key);
int find_hash_table(int hashval, int key);

void
init_hash_table(void)
{
	int i, j;

   /*
    * Initialize Hash Table
    *
    * Set Hash Value for each entry
    * Clear all entries
    */
   for (i=0; i < 10; i++)
   { /* loop through each hash table */
   	HashTable[i].hash_val = i;			/* hash value */
      for (j=0; j < MAX_ENTS; j++)
      { /* loop through each hash table entry */
      	HashTable[i].entry[j].use = USE_NOT;	/* not used */
         HashTable[i].entry[j].key = 0;		/* no key */
         HashTable[i].entry[j].desc[0] = '\0';	/* no desc */
      }
   }
}


int
do_hash_table(void)
{
	char buf[BUFSIZ];

   while (1)
   {
   	printf("\nOptions are:\n");
      printf("A. Add entries to hash_table\n");
      printf("D. Delete entries from hash_table\n");
      printf("L. List hash_table\n");
      printf("S. Search hash_table\n");
      printf("M. Return to Main Menu\n");
      printf("Q. Quit\n");
      printf("Enter option: ");
      gets(buf);
      switch(buf[0])
      {
      case 'a':
      case 'A':
      	add_hash_table();
         break;
      case 'd':
      case 'D':
      	delete_hash_table();
         break;
      case 'l':
      case 'L':
      	list_hash_table();
         break;
      case 's':
      case 'S':
      	search_hash_table();
         break;
      case 'm':
      case 'M':
      	return 1;	/* keep running */

      case 'q':
      case 'Q':
      	return 0;	/* stop running */

      default:
      	printf("Invalid option - try again!\n");
      }
   }
}

/*
 * ADD HASH TABLE
 *
 * This routine adds an entry into the hash table
 * Based on the key, it determines which table (bucket)
 * should receive the new entry
 *
 * It then searches for the first entry that is not in use
 * (i.e., either empty or deleted)
 *
 * It then sets up this new entry
 */

void
add_hash_table(void)
{
	char buf[BUFSIZ];
   int key, hashval, i;

   printf("Enter key: ");
   gets(buf);
   key = atoi(buf);
   hashval = get_hashval(key);
   printf("Hash value of key %d is %d\n", key, hashval);
	printf("Enter description: ");
   gets(buf);

   /* search appropriate hash table bucket for empty slot */
   printf("Searching for available slot in Hash Table %d\n",
   	hashval);
   for (i=0; i < MAX_ENTS; i++)
   {
   	/* if entry in use, skip it */
   	if (HashTable[hashval].entry[i].use == USE_YES) continue;
      /* entry is either empty or delete - use it */
      printf("Slot %d is available, use flag %d (%s)\n",
      	i, HashTable[hashval].entry[i].use,
         get_use_desc(HashTable[hashval].entry[i].use));
      HashTable[hashval].entry[i].use = USE_YES;
      HashTable[hashval].entry[i].key = key;
      strcpy(HashTable[hashval].entry[i].desc, buf);
      return;	/* done */
   }

	/* if we get here, the table was full */
   printf("Hash Table full - delete some first\n");
}

/*
 * DELETE HASH TABLE
 *
 * This routine deletes an entry from the hash table
 *
 * It gets the key from the user, and then searches
 * for that entry.  If not found, it quits
 *
 * To delete the entry, it merely marks the entry as deleted
 */
void
delete_hash_table(void)
{
	char buf[BUFSIZ];
   int key, hashval, ent;

   printf("Enter key to delete: ");
   gets(buf);
   key = atoi(buf);
   hashval = get_hashval(key);
   printf("Hash value for key %d is %d\n", key, hashval);

   /* try to find entry */
   ent = find_hash_table(hashval, key);
   if (ent < 0)
   {
   	printf("Entry for key %d not found.\n", key);
      return;
   }

   /* entry found - check with user */
   printf("Entry to be deleted:\n");
   display_hash_table(hashval, ent);
   printf("OK? (y or n): ");
   gets(buf);
   if (buf[0] != 'y')
   {
   	printf("Entry not deleted\n");
      return;
   }

   /* delete entry */
   printf("Deleting entry %d from Hash Table %d\n",
   	ent, hashval);
   HashTable[hashval].entry[ent].use = USE_DEL;
}

/*
 * LIST HASH TABLE
 *
 * This routine will display the contents of a hash table
 *
 * It gets the hash value from the user to determine which
 * table to display
 */
void
list_hash_table(void)
{
	char buf[BUFSIZ];
   int i, hashval, num_found=0;

   printf("What Hash Table (hashval) do you want to search: ");
   gets(buf);
   hashval = atoi(buf);
	if (hashval < 0 || hashval > 10)
   {
   	printf("Hash value %d is invalid\n", hashval);
      return;
   }

	/* valid value, display appropriate table */
	for (i=0; i < MAX_ENTS; i++)
   {
   	/* quit when we get to empty entry */
   	if (HashTable[hashval].entry[i].use == USE_NOT) break;
      display_hash_table(hashval, i);
      num_found++;
   }
   printf("%d entries found\n", num_found);
}

/*
 * SEARCH HASH TABLE
 *
 * This routine searches for an entry the user specifies
 * most of the work is done by other routines
 */
void
search_hash_table(void)
{
	char buf[BUFSIZ];
   int key, hashval, ent;

   printf("Enter key: ");
   gets(buf);
   key = atoi(buf);
   hashval = get_hashval(key);
   printf("Hash value for key %d is %d\n", key, hashval);

   /* this routine returns index into table */
   ent = find_hash_table(hashval, key);
   if (ent < 0)
   { /* negative indicates entry not found */
   	printf("Entry for key %d not found\n", key);
      return;
   }
   display_hash_table(hashval, ent);
}

/*
 * DISPLAY HASH TABLE
 *
 * This routine displays an entry in a hash table
 * The arguments is the hash value (which indicates
 * which table is to be used), and the index into
 * that table.
 */
void
display_hash_table(int hashval, int ent)
{
	printf("HashVal: %d, Entry %d, use: %d, key: %d, desc: %s\n",
   	hashval, ent, HashTable[hashval].entry[ent].use,
      HashTable[hashval].entry[ent].key,
      HashTable[hashval].entry[ent].desc);
}

/*
 * FIND HASH TABLE
 *
 * This routine searches a hash table looking for an
 * entry containing a matching key
 *
 * The hashval argument indicates which table is to
 * be searched
 *
 * They key indicates which entry to find
 *
 * The routine searches until it finds a match, or
 * it encounters an unused entry
 *
 * There is no need to search past an unused entry
 * since an add will use the first unused or deleted
 * entry.  This means that it couldn't be past an unused
 * entry
 */
int
find_hash_table(int hashval, int key)
{ /* find key in hash table for hashval */
	int i;

   for (i=0; i < MAX_ENTS; i++)
   {
   	/* stop if we find an empty slot */
   	if (HashTable[hashval].entry[i].use == USE_NOT) break;
   	if (HashTable[hashval].entry[i].use == USE_YES
      	&& HashTable[hashval].entry[i].key == key) return i;
   }
   return -1;	/* not found */
}

/*
 * GET HASH VALUE
 *
 * This routine computes the hash value for a given key
 *
 * The hash value is computed by using the last digit of
 * the key (i.e., simply doing a mod 10 on the key
 */
int
get_hashval(int key)
{ /* calculate hash value from key */
   return (key % 10);
}

